# SPDX-FileCopyrightText: 2024 Luca Weiss
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations
from collections.abc import Iterator
from dataclasses import dataclass
from enum import Enum
import json
import random
import shutil
import subprocess
import time

# TODO ESimNotification + list + process + remove


class ESimError(Exception):
    """eSIM managament exception"""
    pass


class ESimProfileState(Enum):
    ENABLED = "enabled"
    DISABLED = "disabled"


@dataclass
class ESimProfile:
    """eSIM profile"""
    _backend: ESimBackend

    iccid: str  # ICCID of Profile
    isdp_aid: str  # Aid of Profile
    state: ESimProfileState  # State of Profile, "Enabled" or "Disabled"
    nickname: str | None  # Nickname of Profile
    service_provider_name: str  # Telecom operators of Profile
    name: str  # Name of Profile
    icon_type: str | None  # Profile icon data struct, "none", "png", "jpg"
    icon: str | None  # Profile icon data in base64
    profile_class: str  # Type of Profile

    def enable(self) -> None:
        self._backend.enable_profile(self.iccid)

    def disable(self) -> None:
        self._backend.disable_profile(self.iccid)

    def delete(self) -> None:
        self._backend.delete_profile(self.iccid)

    def set_nickname(self, nickname: str) -> None:
        self._backend.set_profile_nickname(self.iccid, nickname)

    @classmethod
    def from_dict(cls, backend: ESimBackend, data) -> ESimProfile:
        return cls(
            _backend=backend,
            iccid=data.get('iccid'),
            isdp_aid=data.get('isdpAid'),
            state=ESimProfileState(data.get('profileState')),
            nickname=data.get('profileNickname'),
            service_provider_name=data.get('serviceProviderName'),
            name=data.get('profileName'),
            icon_type=data.get('iconType'),
            icon=data.get('icon'),
            profile_class=data.get('profileClass'),
        )


class ESimManager:
    """eSIM manager"""

    def __init__(self, backends: list[ESimBackend]):
        self._backends: dict[int, ESimBackend] = {}
        for backend in backends:
            slot = backend.get_slot()
            if slot in self._backends:
                raise RuntimeError(f"Cannot have two backends with the same slot '{slot}'")
            self._backends[slot] = backend

    def get_chip_info(self, slot: int) -> dict:
        """Returns various info about the eSIM chip"""
        return self._backends[slot].get_chip_info()

    def get_profiles(self, slot: int) -> list[ESimProfile]:
        """Return the list of profiles on the eSIM"""
        return self._backends[slot].get_profiles()

    # TODO: Make return type somewhat generic
    def download_profile(self, slot: int, smdp_server: str, matching_id: str) -> Iterator[LpacBackend.LpacData | LpacBackend.LpacProgress]:
        """Download a profile onto the eSIM"""
        return self._backends[slot].download_profile(smdp_server, matching_id)

    # TODO: Make return type somewhat generic
    def download_profile_code(self, slot: int, activation_code: str) -> Iterator[LpacBackend.LpacData | LpacBackend.LpacProgress]:
        """Download a profile onto the eSIM"""
        return self._backends[slot].download_profile_code(activation_code)


class ESimBackend:
    """Interface for eSIM backend"""

    def is_operational(self) -> bool:
        """Self-check whether the backend would work on the selected slot"""
        try:
            self.get_chip_info()
            return True
        except Exception:
            return False

    def get_slot(self) -> int:
        """Return the slot number of this backend"""
        raise NotImplementedError

    def get_profiles(self) -> list[ESimProfile]:
        """Return the list of profiles on the eSIM"""
        raise NotImplementedError

    # TODO: Make return type somewhat generic
    def download_profile(self, smdp_server: str, matching_id: str) -> Iterator[LpacBackend.LpacData | LpacBackend.LpacProgress]:
        """Download a profile onto the eSIM"""
        raise NotImplementedError

    # TODO: Make return type somewhat generic
    def download_profile_code(self, activation_code: str) -> Iterator[LpacBackend.LpacData | LpacBackend.LpacProgress]:
        """Download a profile onto the eSIM"""
        raise NotImplementedError

    def enable_profile(self, iccid: str) -> None:
        """Enable the profile with the given iccid"""
        raise NotImplementedError

    def disable_profile(self, iccid: str) -> None:
        """Disable the profile with the given iccid"""
        raise NotImplementedError

    def delete_profile(self, iccid: str) -> None:
        """Delete the profile with the given iccid"""
        raise NotImplementedError

    def set_profile_nickname(self, iccid: str, nickname: str) -> None:
        """Set the nickname for the profile with the given iccid"""
        raise NotImplementedError

    def get_chip_info(self) -> dict:
        """Returns various info about the eSIM chip"""
        raise NotImplementedError


class LpacBackend(ESimBackend):
    """Implementation of ESimBackend using lpac"""

    @dataclass
    class LpacData:
        data: dict | None

    @dataclass
    class LpacProgress:
        message: str
        data: dict | None

    def __init__(self, slot: int):
        """Initialize the lpac eSIM backend with the given slot number"""
        self.slot = slot
        self.lpac_exe = shutil.which("lpac")
        if not self.lpac_exe:
            raise ESimError("Required executable 'lpac' not found.")

    def __parse_output(self, output: dict) -> LpacBackend.LpacData | LpacBackend.LpacProgress:
        data_type = output["type"]
        payload = output["payload"]

        if data_type == "lpa":
            code = payload["code"]
            if code != 0:
                raise ESimError(f"Error from lpac: {code}, message={payload["message"]} data={payload["data"]}")
            return LpacBackend.LpacData(data=payload["data"])
        if data_type == "progress":
            code = payload["code"]
            if code != 0:
                raise ESimError(f"Error from lpac: {code}, message={payload["message"]} data={payload["data"]}")
            return LpacBackend.LpacProgress(message=payload["message"], data=payload["data"])

        raise ESimError(f"Unknown lpac output type {data_type}")

    def __run_lpac(self, lpac_args) -> dict:
        env = {"LPAC_APDU": "qmi_qrtr", "UIM_SLOT": str(self.slot)}
        output = subprocess.check_output([self.lpac_exe] + lpac_args, env=env)
        res = self.__parse_output(json.loads(output))
        if not isinstance(res, LpacBackend.LpacData):
            raise ESimError(f"Unexpected data: {res}")
        return res.data

    def __run_lpac_stream(self, lpac_args) -> Iterator[LpacBackend.LpacData | LpacBackend.LpacProgress]:
        env = {"LPAC_APDU": "qmi_qrtr", "UIM_SLOT": str(self.slot)}
        with subprocess.Popen([self.lpac_exe] + lpac_args, env=env, stdout=subprocess.PIPE, stderr=subprocess.STDOUT) as process:
            for line_obj in process.stdout:
                line = line_obj.decode('utf8')
                try:
                    obj = json.loads(line)
                    yield self.__parse_output(obj)
                except json.decoder.JSONDecodeError:
                    # TODO Also propagate this through the stack somehow? New LpacMisc class?
                    print(f"Got non-JSON line, ignoring: {line.strip()}")

        if process.returncode != 0:
            raise ESimError(f"Unexpected lpac exit code: {process.returncode}")

    def get_slot(self) -> int:
        return self.slot

    def get_profiles(self) -> list[ESimProfile]:
        profiles = []
        for profile_data in self.__run_lpac(["profile", "list"]):
            profile = ESimProfile.from_dict(self, profile_data)
            profiles.append(profile)

        return profiles

    def download_profile(self, smdp_server: str, matching_id: str) -> Iterator[LpacBackend.LpacData | LpacBackend.LpacProgress]:
        res = self.__run_lpac_stream(["profile", "download", "-s", smdp_server, "-m", matching_id])
        return res

    def download_profile_code(self, activation_code: str) -> Iterator[LpacBackend.LpacData | LpacBackend.LpacProgress]:
        res = self.__run_lpac_stream(["profile", "download", "-a", activation_code])
        return res

    def enable_profile(self, iccid: str) -> None:
        self.__run_lpac(["profile", "enable", iccid])

    def disable_profile(self, iccid: str) -> None:
        self.__run_lpac(["profile", "disable", iccid])

    def delete_profile(self, iccid: str) -> None:
        self.__run_lpac(["profile", "delete", iccid])

    def set_profile_nickname(self, iccid: str, nickname: str) -> None:
        self.__run_lpac(["profile", "nickname", iccid, nickname])

    def get_chip_info(self) -> dict:
        return self.__run_lpac(["chip", "info"])


class DummyBackend(ESimBackend):
    """Implementation of ESimBackend for testing purposes"""

    def __init__(self, slot: int):
        self.slot = slot

        self.profiles = [
            ESimProfile(
                _backend=self,
                iccid="89001010001234567063",
                isdp_aid="a0000005591010ffffffff8900ffff00",
                state=ESimProfileState.DISABLED,
                nickname=None,
                service_provider_name="MyMNO1",
                name="MyMNO1Prof",
                icon_type=None,
                icon=None,
                profile_class="operational",
            )
        ]

    def _find_profile(self, iccid: str) -> ESimProfile:
        for profile in self.profiles:
            if profile.iccid == iccid:
                return profile
        raise IndexError(f"Profile with ICCID {iccid} not found")

    def get_slot(self) -> int:
        return self.slot

    def get_profiles(self) -> list[ESimProfile]:
        return self.profiles

    def download_profile(self, smdp_server: str, matching_id: str) -> Iterator[LpacBackend.LpacData | LpacBackend.LpacProgress]:
        yield LpacBackend.LpacProgress("es10b_get_euicc_challenge_and_info", None)
        time.sleep(0.5)
        yield LpacBackend.LpacProgress("es9p_initiate_authentication", None)
        time.sleep(0.5)
        yield LpacBackend.LpacProgress("es10b_authenticate_server", None)
        time.sleep(0.5)
        yield LpacBackend.LpacProgress("es9p_authenticate_client", None)
        time.sleep(0.5)
        yield LpacBackend.LpacProgress("es10b_prepare_download", None)
        time.sleep(0.5)
        yield LpacBackend.LpacProgress("es9p_get_bound_profile_package", None)
        time.sleep(0.5)
        yield LpacBackend.LpacProgress("es10b_load_bound_profile_package", None)
        time.sleep(0.5)
        yield LpacBackend.LpacData(None)
        dummy_nr = random.randint(1000, 9999)
        self.profiles.append(
            ESimProfile(
                _backend=self,
                iccid=f"8900101000123456{dummy_nr}",
                isdp_aid="a0000005591010ffffffff8900ffff01",
                state=ESimProfileState.DISABLED,
                nickname=None,
                service_provider_name=f"Dummy Provider {dummy_nr}",
                name=f"Dummy Profile {dummy_nr}",
                icon_type=None,
                icon=None,
                profile_class="operational",
            )
        )

    def download_profile_code(self, activation_code: str) -> Iterator[LpacBackend.LpacData | LpacBackend.LpacProgress]:
        return self.download_profile("foo", "bar")

    def enable_profile(self, iccid: str) -> None:
        profile = self._find_profile(iccid)
        profile.state = ESimProfileState.ENABLED

    def disable_profile(self, iccid: str) -> None:
        profile = self._find_profile(iccid)
        profile.state = ESimProfileState.DISABLED

    def delete_profile(self, iccid: str) -> None:
        profile = self._find_profile(iccid)
        self.profiles.remove(profile)

    def set_profile_nickname(self, iccid: str, nickname: str | None) -> None:
        profile = self._find_profile(iccid)
        profile.nickname = nickname

    def get_chip_info(self) -> dict:
        # Return completely bogus information but in a similar format with nesting etc
        return {
            "key": "value",
            "child_obj": {
                "key2": "value2",
            },
            "child_array_string": [
                "foo", "bar"
            ],
            "child_array_object": [
                {"key": "value"},
                {"key_int": 99},
                {"key_null": None},
            ],
        }


# TODO: Remove, testing code
def main():
    from pprint import pprint
    backend = LpacBackend(2)
    # backend = DummyBackend()
    manager = ESimManager([backend])
    pprint(manager.get_profiles(2))

    pprint(manager.get_chip_info(2))

    for val in manager.download_profile(2, "rsp-eu.example.org", "1234567890ABCDEF1234567890ABCDEF"):
        if isinstance(val, LpacBackend.LpacProgress):
            print(f"Got progress! {val.message} - {val.data}")
        elif isinstance(val, LpacBackend.LpacData):
            print(f"Got final data: {val}")
        else:
            raise RuntimeError(f"Unexpected val: {val}")


if __name__ == '__main__':
    main()
